@@ -1,3 +1,35 @@
+# BSD 3-Clause License
+#
+# Copyright (c) 2024, ddelange, <ddelange@delange.dev>
+#
+# All rights reserved.
+#
+# Redistribution and use in source and binary forms, with or without
+# modification, are permitted provided that the following conditions are met:
+#
+# 1. Redistributions of source code must retain the above copyright notice, this
+#    list of conditions and the following disclaimer.
+#
+# 2. Redistributions in binary form must reproduce the above copyright notice,
+#    this list of conditions and the following disclaimer in the documentation
+#    and/or other materials provided with the distribution.
+#
+# 3. Neither the name of the copyright holder nor the names of its
+#    contributors may be used to endorse or promote products derived from
+#    this software without specific prior written permission.
+#
+# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
+# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
+# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
+# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+#
+# SPDX-License-Identifier: BSD-3-Clause
 """Submodule containing code to run PandasObject.apply() in parallel.
 
 Standalone usage (without init):
@@ -10,17 +42,35 @@ Standalone usage (without init):
 
     df["squared"] = mapply(df.A, lambda x: x ** 2, progressbar=False)
 """
-from functools import partial
-from typing import Any, Callable, Tuple, Union
 
+from __future__ import annotations
+
+import warnings
+from functools import partial
+from typing import Any, Callable
+
+from mapply._groupby import run_groupwise_apply
 from mapply.parallel import N_CORES, multiprocessing_imap
 
 DEFAULT_CHUNK_SIZE = 100
 DEFAULT_MAX_CHUNKS_PER_WORKER = 8
 
 
+warnings.filterwarnings(
+    action="ignore",
+    message=".*swapaxes",
+    category=FutureWarning,
+)
+
+warnings.filterwarnings(
+    action="ignore",
+    message=".*grouper",
+    category=FutureWarning,
+)
+
+
 def _choose_n_chunks(
-    shape: Tuple[int, ...],
+    shape: tuple[int, ...],
     axis: int,
     n_workers: int,
     chunk_size: int,
@@ -29,10 +79,12 @@ def _choose_n_chunks(
     """Choose final amount of chunks to be sent to the ProcessPool."""
     # no sense running parallel if data is too small
     n_chunks = int(shape[axis] / chunk_size)
+    if n_workers < 1:
+        n_workers = N_CORES
 
     if max_chunks_per_worker:
         # no sense making too many chunks
-        n_chunks = min(n_chunks, max_chunks_per_worker * N_CORES)
+        n_chunks = min(n_chunks, max_chunks_per_worker * n_workers)
     if n_chunks < 1 or n_workers == 1 or N_CORES == 1:
         # no sense running parallel
         n_chunks = 1
@@ -40,22 +92,22 @@ def _choose_n_chunks(
     return n_chunks
 
 
-def mapply(
+def mapply(  # noqa: PLR0913
     df_or_series: Any,
     func: Callable,
-    axis: Union[int, str] = 0,
+    axis: int | str = 0,
     *,
     n_workers: int = -1,
     chunk_size: int = DEFAULT_CHUNK_SIZE,
     max_chunks_per_worker: int = DEFAULT_MAX_CHUNKS_PER_WORKER,
     progressbar: bool = True,
-    args=(),
-    **kwargs
+    args: tuple[Any, ...] = (),
+    **kwargs: Any,
 ) -> Any:
     """Run apply on n_workers. Split in chunks if sensible, gather results, and concat.
 
     When using :meth:`mapply.init`, the signature of this method will behave the same as
-    :meth:`pandas.DataFrame.apply` or :meth:`pandas.Series.apply`.
+    :meth:`pandas.DataFrame.apply`/:meth:`pandas.Series.apply`/:meth:`pandas.core.groupby.GroupBy.apply`.
 
     Args:
         df_or_series: Argument reserved to the class instance, a.k.a. 'self'.
@@ -70,7 +122,7 @@ def mapply(
             n_chunks determined by chunk_size if necessary. Set to 0 to skip this check.
         progressbar: Whether to wrap the chunks in a :meth:`tqdm.auto.tqdm`.
         args: Additional positional arguments to pass to func.
-        kwargs: Additional keyword arguments to pass to apply/func.
+        **kwargs: Additional keyword arguments to pass to apply/func.
 
     Returns:
         Series or DataFrame resulting from applying func along given axis.
@@ -80,6 +132,17 @@ def mapply(
     """
     from numpy import array_split
     from pandas import Series, concat
+    from pandas.core.groupby import GroupBy
+
+    if isinstance(df_or_series, GroupBy):
+        return run_groupwise_apply(
+            df_or_series,
+            func,
+            n_workers=n_workers,
+            progressbar=progressbar,
+            args=args,
+            **kwargs,
+        )
 
     if isinstance(axis, str):
         axis = ["index", "columns"].index(axis)
@@ -87,7 +150,8 @@ def mapply(
     isseries = int(isinstance(df_or_series, Series))
 
     if isseries and axis == 1:
-        raise ValueError("Passing axis=1 is not allowed for Series")
+        msg = "Passing axis=1 is not allowed for Series"
+        raise ValueError(msg)
 
     opposite_axis = 1 - (isseries or axis)
 
@@ -113,7 +177,7 @@ def mapply(
             dfs,
             n_workers=n_workers,
             progressbar=progressbar,
-        )
+        ),
     )
 
     if isseries or len(results) == 1 or sum(map(len, results)) in df_or_series.shape:
